from __future__ import annotations

import json
import re
from dataclasses import dataclass, field

from typing_extensions import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from scrapy_webarchive.wacz.wacz_file import WaczFile

CDXREC = re.compile(
    r"^(?P<surt>(?P<host>[^\)\s]+)\)(?P<path>[^\?\s]+)?(\?(?P<query>\S+))?)"
    r"\s(?P<datetime>(?P<year>\d{4})(?P<month>\d{2})(?P<day>\d{2})(?P<hour>\d{2})(?P<minute>\d{2})(?P<second>\d{2})(?:\d{3})?)"
    r"\s(?P<data>{.*})"
)


@dataclass
class CdxjRecord:
    """Represents a CDXJ record, which contains metadata about archived web content stored in WARC files."""

    wacz_file: "WaczFile"
    surt: str
    host: str
    path: str = ""
    query: str = ""
    datetime: str = ""
    year: str = ""
    month: str = ""
    day: str = ""
    hour: str = ""
    minute: str = ""
    second: str = ""
    data: dict = field(default_factory=dict)

    @staticmethod
    def _parse(line: str):
        """Parses a single line from a CDXJ index."""

        return CDXREC.match(line)

    @classmethod
    def from_cdxline(cls, cdxline: str, wacz_file: "WaczFile") -> Union[CdxjRecord, None]:
        """Creates a CdxjRecord instance from a CDX(J) line."""

        m = cls._parse(cdxline.strip())

        if not m:
            return None

        parsed_data = m.groupdict(default="")
        parsed_data['data'] = json.loads(parsed_data['data'])

        return cls(**parsed_data, wacz_file=wacz_file)

    def __str__(self):
        return str(self.__dict__)



from __future__ import annotations

import re
from datetime import datetime
from urllib.parse import urlparse

from scrapy import Request, Spider, signals
from scrapy.crawler import Crawler
from scrapy.exceptions import NotConfigured
from scrapy.settings import Settings
from scrapy.statscollectors import StatsCollector
from typing_extensions import Iterable, List, Optional, Self, Union

from scrapy_webarchive import utils
from scrapy_webarchive.exceptions import WaczMiddlewareException
from scrapy_webarchive.resolvers import create_resolver
from scrapy_webarchive.strategies import FileLookupStrategy, StrategyRegistry
from scrapy_webarchive.wacz.storages import ZipStorageHandlerFactory
from scrapy_webarchive.wacz.wacz_file import MultiWaczFile, WaczFile
from scrapy_webarchive.warc.transformers import record_transformer


class BaseWaczMiddleware:
    """A base class for middlewares that require opening one or more WACZ files."""

    wacz: Union[WaczFile, MultiWaczFile]

    def __init__(self, settings: Settings, stats: StatsCollector, spider_name) -> None:
        self.stats = stats
        self.settings = settings
        self.spider_name = spider_name
        self.wacz_uris = self._resolve_wacz_uris()

        if not self.wacz_uris:
            raise NotConfigured

        self.crawl = settings.get("SW_WACZ_CRAWL", False)

    @classmethod
    def from_crawler(cls, crawler: Crawler) -> Self:
        assert crawler.stats
        spider_name = crawler.spidercls.name if hasattr(crawler.spidercls, "name") else crawler.spider.name
        o = cls(crawler.settings, crawler.stats, spider_name)
        crawler.signals.connect(o.spider_opened, signals.spider_opened)
        return o

    def spider_opened(self, spider: Spider) -> None:
        """
        Handles the initialization of WACZ files when the spider is opened.

        This method is called when the Scrapy spider starts, and it attempts to open WACZ files 
        defined in the spider's configuration. It processes multiple WACZ URIs provided in `self.wacz_uris`, logs the 
        process, and collects valid WACZ files for further use.

        If only one WACZ URI is provided, it opens and assigns the file to `self.wacz` as a `WaczFile` instance. 
        If multiple URIs are provided, valid files are assigned to `self.wacz` as a `MultiWaczFile` instance.
        """

        spider.logger.info(f"[WACZDownloader] Found {len(self.wacz_uris)} WACZ URI(s) to open")
        wacz_files = []
        
        for wacz_uri in self.wacz_uris:
            spider.logger.info(f"[WACZDownloader] Opening WACZ {wacz_uri}")
            storage_handler = ZipStorageHandlerFactory.get_handler(wacz_uri, spider.settings)

            if not storage_handler.zip_exists:
                spider.logger.error(f"[WACZDownloader] Could not open WACZ {wacz_uri}")
                continue

            wacz_files.append(WaczFile(storage_handler=storage_handler))

        if wacz_files:
            spider.logger.info(
                f"[WACZDownloader] Continuing with {len(wacz_files)}/{len(self.wacz_uris)} valid WACZ files"
            )
            self.wacz = wacz_files[0] if len(wacz_files) == 1 else MultiWaczFile(wacz_files)

        # If there are not wacz_files, we raise a `WaczMiddlewareException` in the downloader/spider middleware.
        # Raising an exception here does not stop the job from running. If there are no valid WACZ files configured
        # we do not want to continue the job.

    def _resolve_wacz_uris(self) -> List[str]:
        """Resolve WACZ URIs based on settings."""
        
        if self.settings.get("SW_WACZ_SOURCE_URI", False):
            return re.split(r"\s*,\s*", self.settings.get("SW_WACZ_SOURCE_URI"))

        if not self._uri_template or not self._target_time or not self._strategy:
            return []

        base_path = utils.extract_base_from_uri_template(self._uri_template)

        # Edge case where the URI template does not contain any placeholders.
        if base_path == self._uri_template and not utils.is_uri_directory(self._uri_template):
            # Export URI does not require scheme. Assume it is a local file path if no scheme is defined.
            if self._uri_template.startswith("/"):
                return [f"file://{self._uri_template}"]
            return [self._uri_template]

        regex_pattern = utils.build_regex_pattern(
            self._uri_template, 
            placeholder_patterns=utils.get_placeholder_patterns(self.spider_name)
        )
        resolver = create_resolver(settings=self.settings, base_path=base_path, regex_pattern=regex_pattern)
        matching_files = resolver.resolve()

        if not matching_files:
            return []

        result = self._strategy.find(files=matching_files, target=self._target_time)
        return [result] if result else []

    def _is_off_site(self, url: str, spider: Spider) -> bool:
        """Check if the URL is off-site based on allowed domains."""

        return hasattr(spider, "allowed_domains") and urlparse(url).hostname not in spider.allowed_domains

    def _is_filtered_by_spider(self, url: str, spider: Spider) -> bool:
        """Check if the URL is filtered out by the spider's archive rules."""

        return hasattr(spider, "archive_regexp") and \
               re.search(spider.archive_regexp, url) is None

    def _is_blacklisted_by_spider(self, url: str, spider: Spider) -> bool:
        """Check if the URL is blacklisted out by the spider's archive blacklist rules."""

        return hasattr(spider, "archive_blacklist_regexp") and \
               re.search(spider.archive_blacklist_regexp, url) is not None

    @property
    def _uri_template(self) -> Optional[str]:
        """Generates the search pattern based on the export URI format."""

        uri_template = self.settings.get("SW_EXPORT_URI")

        if not uri_template:
            return None
        
        if utils.is_uri_directory(uri_template):
            uri_template += "{filename}"

        return uri_template

    @property
    def _target_time(self) -> Optional[datetime]:
        """Parse the lookup target into a datetime object."""

        lookup_target = self.settings.get("SW_WACZ_LOOKUP_TARGET")
        return utils.parse_iso8601_datetime(lookup_target)

    @property
    def _strategy(self) -> FileLookupStrategy:
        """Get the lookup strategy from the registry."""

        return StrategyRegistry.get(self.settings.get("SW_WACZ_LOOKUP_STRATEGY", "after"))


class WaczCrawlMiddleware(BaseWaczMiddleware):
    """
    Scrapy WACZ crawl spider middleware to crawl from a WACZ archive.

    Replaces the default behaviour of the spider by solely iterating over the
    entries in the WACZ file.
    """

    def spider_opened(self, spider: Spider) -> None:
        if not self.crawl:
            return

        super().spider_opened(spider)

    def process_start_requests(self, start_requests: Iterable[Request], spider: Spider) -> Iterable[Request]:
        """Processes start requests and yields WACZ index entries or original requests based on the crawl setting."""
        
        # If crawl is disabled, yield the original start requests.
        if not self.crawl:
            yield from start_requests
            return
        # If the attribute has not been set, none of the WACZ could be opened.
        elif not hasattr(self, "wacz"):
            raise WaczMiddlewareException("Could not open any WACZ files, check your WACZ URIs and authentication.")

        # Iterate over entries in the WACZ index.
        for entry in self.wacz.iter_index():
            url = entry.data["url"]

            # Filter out off-site requests
            if self._is_off_site(url, spider):
                self.stats.inc_value("webarchive/crawl_skip/off_site", spider=spider)
            # Ignore filtered pages (to avoid crawling all files in the WACZ)
            elif self._is_filtered_by_spider(url, spider):
                self.stats.inc_value("webarchive/crawl_skip/filtered", spider=spider)
            # Also ignore blacklisted pages (as it would be illogical to return them here, though use-case is different)
            elif self._is_blacklisted_by_spider(url, spider):
                self.stats.inc_value("webarchive/crawl_skip/blacklisted", spider=spider)
            else:
                self.stats.inc_value("webarchive/start_request_count", spider=spider)
                flags = ["wacz_start_request"]

                yield record_transformer.request_for_record(
                    entry,
                    flags=flags,
                    meta={"cdxj_record": entry},
                    dont_filter=True,
                )




from scrapy_webarchive.cdxj.models import CdxjRecord


def test_cdxj_record_valid():
    # Sample valid CDXJ line
    valid_cdxj_line = "com,example)/index 20241003000000 {\"url\": \"http://example.com/index\", \"status\": \"200\"}"
    
    # Create a CdxjRecord object
    record = CdxjRecord.from_cdxline(valid_cdxj_line, wacz_file=None)
    
    # Test extracted data from the CDXJ line
    assert record.surt == "com,example)/index"
    assert record.host == "com,example"
    assert record.path == "/index"
    assert record.year == "2024"
    assert record.month == "10"
    assert record.day == "03"
    assert record.data == {"url": "http://example.com/index", "status": "200"}
    assert record.datetime == "20241003000000"
    assert str(record)


def test_cdxj_record_invalid_format():
    # Invalid CDXJ line (missing date)
    invalid_cdxj_line = "com,example)/index {\"url\": \"http://example.com/index\", \"status\": \"200\"}"
    
    # Test that the invalid line raises returns None
    assert CdxjRecord.from_cdxline(invalid_cdxj_line, wacz_file=None) is None


def test_cdxj_record_invalid_json_data():
    # Invalid JSON in CDXJ line
    invalid_cdxj_line = "com,example)/index 20241003000000 {\"url\": \"http://example.com/index\", \"status\": \"200\""
    
    # Test that the invalid JSON returns None
    assert CdxjRecord.from_cdxline(invalid_cdxj_line, wacz_file=None) is None


def test_cdxj_record_empty_line():
    # Test that an empty line returns None
    assert CdxjRecord.from_cdxline('', wacz_file=None) is None


def test_cdxj_record_no_data_field():
    # CDXJ line with no data field
    no_data_cdxj_line = "com,example)/index 20241003000000"
    
    # Test that no data field returns None
    assert CdxjRecord.from_cdxline(no_data_cdxj_line, wacz_file=None) is None


def test_cdxj_record_urn_pageinfo():
    # CDXJ line starting with urn:pageinfo
    url = "urn:pageinfo:https://example.com/index"
    pageinfo_cdxj_line = url + " 20241003000000 {\"url\": \"" + url + "\" }"

    # Test that the urn:pageinfo line returns None
    assert CdxjRecord.from_cdxline(pageinfo_cdxj_line, wacz_file=None) is None



from __future__ import annotations

import hashlib
import logging
import re
from datetime import datetime, timezone
from pathlib import Path
from urllib.parse import urlparse

from typing_extensions import IO, Dict, Optional, Tuple

from scrapy_webarchive.constants import BUFF_SIZE

logger = logging.getLogger(__name__)


def get_formatted_dt_string(format: str) -> str:
    return datetime.now(timezone.utc).strftime(format)


def header_lines_to_dict(lines):
    # TODO: Only supports each header appearing once, not multiple occurences.
    headers = {}
    for line in lines:
        k, v = line.split(b":", 1)
        v = v.lstrip()
        headers[k] = v
    return headers


def get_scheme_from_uri(uri: str) -> str:
    if Path(uri).is_absolute():  # to support win32 paths like: C:\\some\dir.
        return "file"
    else:
        return urlparse(uri).scheme


def hash_stream(hash_type: str, stream: IO) -> Tuple[int, str]:
    """Hashes the stream with given hash_type hasher."""

    # At this moment the `hash_type` (or algorithm) that we pass will always be sha256 as it is hardcoded.
    # This check is implemented in case any other algorithms will be made available in the future.
    if hash_type not in hashlib.algorithms_guaranteed:
        raise ValueError(f"Unsupported hash type: {hash_type}")

    hasher = hashlib.new(hash_type)

    size = 0
    for chunk in iter(lambda: stream.read(BUFF_SIZE), b""):
        size += len(chunk)
        hasher.update(chunk)

    return size, f"{hash_type}:{hasher.hexdigest()}"


def parse_iso8601_datetime(time_str: str) -> Optional[datetime]:
    """Convert the target ISO time string to a datetime object."""

    if not time_str:
        return None

    try:
        return datetime.fromisoformat(time_str)
    except ValueError:
        raise ValueError(f"Invalid date format: {time_str}. Use ISO format (YYYY-MM-DDTHH:MM:SS).")


def get_archive_uri_template_dt_variables() -> dict:
    current_date = datetime.now()

    return {
        "year": current_date.strftime("%Y"),
        "month": current_date.strftime("%m"),
        "day": current_date.strftime("%d"),
        "timestamp": current_date.strftime("%Y%m%d%H%M%S"),
    }


def get_placeholder_patterns(spider_name: str) -> Dict[str, str]:
    """Return a mapping of placeholders to their corresponding regex patterns."""

    return {
        "{year}": r"[0-9]{4}",
        "{month}": r"[0-9]{2}",
        "{day}": r"[0-9]{2}",
        "{timestamp}": r"[0-9]+",
        "{spider}": spider_name,
        "{filename}": r"[^/\\]+\.wacz$",
    }


def is_uri_directory(uri: str) -> bool:
    """Check if the URI is a directory or a file."""

    parsed = urlparse(uri)
    scheme = parsed.scheme
    path = parsed.path

    if not scheme and uri.startswith("/"):
        path = uri

    last_part = path.rstrip("/").split("/")[-1]

    if path.endswith("/"):
        return True
    elif "." in last_part:
        return False
    else:
        return True


def extract_base_from_uri_template(uri_template: str) -> str:
    """Extract the static base path from a URI template before the first placeholder."""

    first_placeholder_pos = uri_template.find("{")
    if first_placeholder_pos == -1:
        return uri_template

    return uri_template[:first_placeholder_pos]


def build_regex_pattern(uri_template: str, placeholder_patterns: Dict[str, str]) -> re.Pattern:
    """Build and compile a regex pattern from a URI template with placeholders."""

    first_placeholder_pos = uri_template.find("{")
    pattern_str = uri_template[first_placeholder_pos:]
    for placeholder, regex in placeholder_patterns.items():
        pattern_str = pattern_str.replace(placeholder, regex)

    return re.compile(pattern_str)



import os
import re
from datetime import datetime
from pathlib import Path
from unittest.mock import Mock, patch

import pytest
from freezegun import freeze_time

from scrapy_webarchive.models import FileInfo
from scrapy_webarchive.resolvers import LocalFileResolver, S3FileResolver, create_resolver


@pytest.fixture
def mock_s3_client():
    return Mock()


@pytest.fixture
def sample_regex():
    return re.compile(r".*\.wacz$")


class TestS3FileResolver:
    def test_resolve_with_matching_files(self, mock_s3_client, sample_regex):
        mock_response = {
            "Contents": [
                {"Key": "archive_1.wacz", "LastModified": datetime(2025, 1, 1)},
                {"Key": "archive_2.warc", "LastModified": datetime(2025, 1, 2)},
                {"Key": "archive_3.wacz", "LastModified": datetime(2025, 1, 3)},
            ]
        }
        mock_s3_client.list_objects_v2.return_value = mock_response

        resolver = S3FileResolver(mock_s3_client, "scrapy-webarchive", sample_regex)
        result = resolver.resolve()

        expected = [
            FileInfo(uri="s3://scrapy-webarchive/archive_1.wacz", last_modified=datetime(2025, 1, 1).timestamp()),
            FileInfo(uri="s3://scrapy-webarchive/archive_3.wacz", last_modified=datetime(2025, 1, 3).timestamp()),
        ]
        
        assert result == expected
        mock_s3_client.list_objects_v2.assert_called_once_with(Bucket="scrapy-webarchive")

    def test_resolve_no_contents(self, mock_s3_client, sample_regex):
        mock_s3_client.list_objects_v2.return_value = {}
        resolver = S3FileResolver(mock_s3_client, "scrapy-webarchive", sample_regex)

        result = resolver.resolve()
        assert result == []


class TestLocalFileResolver:
    @freeze_time("2025-01-01 12:00:00")
    def test_resolve_with_matching_files(self, sample_regex, fs):
        base_path = "/scrapy-webarchive"
        fs.create_file(os.path.join(base_path, "archive_1.wacz"), contents="")
        fs.create_file(os.path.join(base_path, "archive_2.warc"), contents="")
        fs.create_file(os.path.join(base_path, "archive_3.wacz"), contents="")

        resolver = LocalFileResolver(base_path, sample_regex)
        result = resolver.resolve()

        expected_last_modified = datetime(2025, 1, 1, 12).timestamp() # frozen time
        expected = [
            FileInfo(uri="file:///scrapy-webarchive/archive_1.wacz", last_modified=expected_last_modified),
            FileInfo(uri="file:///scrapy-webarchive/archive_3.wacz", last_modified=expected_last_modified),
        ]
        assert result == expected


class TestCreateResolver:
    @patch('scrapy_webarchive.resolvers.get_s3_client')
    def test_create_s3_resolver(self, mock_get_s3_client, sample_regex):
        mock_s3_client = Mock()
        mock_get_s3_client.return_value = mock_s3_client

        resolver = create_resolver({}, "s3://scrapy-webarchive/quotes/", sample_regex)

        assert isinstance(resolver, S3FileResolver)
        assert resolver.bucket == "scrapy-webarchive"
        assert resolver.s3_client == mock_s3_client
        assert resolver.pattern_regex.pattern == "quotes/.*\\.wacz$"

    def test_create_local_resolver(self, sample_regex):
        resolver = create_resolver({}, "/scrapy-webarchive", sample_regex)

        assert isinstance(resolver, LocalFileResolver)
        assert resolver.base_path == Path("/scrapy-webarchive")
        assert resolver.pattern_regex == sample_regex



import hashlib
import io
import re

import pytest

from scrapy_webarchive import utils
from scrapy_webarchive.constants import BUFF_SIZE


def test_hash_stream_with_empty_stream():
    data = b""
    stream = io.BytesIO(data)
    size, result = utils.hash_stream("sha256", stream)
    
    assert size == 0
    assert result == f"sha256:{hashlib.sha256(data).hexdigest()}"


def test_hash_stream_with_md5_algorithm():
    data = b"Hello world"
    expected_hash = hashlib.md5(data).hexdigest()
    
    stream = io.BytesIO(data)
    size, result = utils.hash_stream("md5", stream)
    
    assert size == len(data)
    assert result == f"md5:{expected_hash}"


def test_hash_stream_with_sha256_algorithm():
    data = b"Hello world"
    expected_hash = hashlib.sha256(data).hexdigest()
    
    stream = io.BytesIO(data)
    size, result = utils.hash_stream("sha256", stream)
    
    assert size == len(data)
    assert result == f"sha256:{expected_hash}"


def test_hash_stream_with_unsupported_hash_type():
    data = b"Hello world"
    stream = io.BytesIO(data)

    with pytest.raises(ValueError):
        utils.hash_stream("unsupported_hash", stream)


def test_hash_stream_with_large_stream():
    data = b"a" * (2 * BUFF_SIZE)  # Twice the buffer size
    expected_hash = hashlib.sha256(data).hexdigest()
    
    stream = io.BytesIO(data)
    size, result = utils.hash_stream("sha256", stream)
    
    assert size == len(data)
    assert result == f"sha256:{expected_hash}"


@pytest.mark.parametrize("uri, expected", [
    # Directories
    ("s3://scrapy-webarchive/quotes/", True),
    ("s3://scrapy-webarchive/quotes", True),
    ("/scrapy/webarchive/", True),
    ("scrapy/webarchive/", True),
    ("/scrapy-webarchive", True),

    # Files
    ("s3://scrapy-webarchive/quotes/archive.wacz", False),
    ("s3://scrapy-webarchive/archive.wacz", False),
    ("/scrapy/webarchive/archive.wacz", False),
])
def test_is_uri_directory(uri: str, expected: bool):
    assert utils.is_uri_directory(uri) == expected


@pytest.mark.parametrize("uri_template, expected", [
    ("s3://scrapy-webarchive/{year}/quotes/{day}/res-{timestamp}.wacz", "s3://scrapy-webarchive/"),
    ("s3://scrapy-webarchive/quotes/{year}/{month}/{day}/res-{timestamp}.wacz", "s3://scrapy-webarchive/quotes/"),
    ("s3://scrapy-webarchive/", "s3://scrapy-webarchive/"),
    ("file:///scrapy/{year}/quotes/{day}/res-{timestamp}.wacz", "file:///scrapy/"),
    ("file:///scrapy/webarchive/{year}/", "file:///scrapy/webarchive/"),
    ("file:///scrapy/webarchive/", "file:///scrapy/webarchive/"),
    ("/scrapy/{year}/quotes/{day}/res-{timestamp}.wacz", "/scrapy/"),
    ("/scrapy/webarchive/{spider}/", "/scrapy/webarchive/"),
    ("/scrapy/webarchive/", "/scrapy/webarchive/"),
])
def test_extract_base_from_uri_template(uri_template: str, expected: str):
    result = utils.extract_base_from_uri_template(uri_template)
    assert result == expected


def test_build_regex_pattern_with_placeholders():
    uri_template = "s3://scrapy-webarchive/{spider}/{year}/{month}/{day}/{timestamp}/"
    result = utils.build_regex_pattern(uri_template, utils.get_placeholder_patterns(spider_name="quotes"))

    assert result.pattern == "quotes/[0-9]{4}/[0-9]{2}/[0-9]{2}/[0-9]+/"
    assert isinstance(result, re.Pattern)

    assert result.match("quotes/2025/01/01/1735686000/")
    assert not result.match("quotes/01/01/2025/1735686000/") # Year is not first



from __future__ import annotations

import socket

from scrapy_webarchive.constants import TIMESTAMP_DT_FORMAT
from scrapy_webarchive.utils import get_formatted_dt_string


def generate_warc_fname(prefix: str) -> str:
    """
    Returns new WARC filename based on recommendation in the warc-specification:
    https://iipc.github.io/warc-specifications/specifications/warc-format/warc-1.1/#annex-c-informative-warc-file-size-and-name-recommendations
    {prefix}-{timestamp}-{serial}-{crawlhost}.warc.gz
    """

    timestamp = get_formatted_dt_string(format=TIMESTAMP_DT_FORMAT)
    crawlhost = socket.gethostname().split(".")[0]
    # As of now we only generate one WARC file. Add serial in here to adhere to the warc specification.
    serial = '00000'

    return "-".join([prefix, timestamp, serial, crawlhost]) + ".warc.gz"



from __future__ import annotations

import re
from pathlib import Path
from urllib.parse import urlparse

from scrapy.settings import Settings
from typing_extensions import List, Protocol

from scrapy_webarchive.models import FileInfo
from scrapy_webarchive.wacz.storages import get_s3_client


class FileResolver(Protocol):
    def resolve(self) -> List[FileInfo]:
        ...


class S3FileResolver:
    def __init__(self, s3_client, bucket: str, regex_pattern: re.Pattern):
        self.s3_client = s3_client
        self.bucket = bucket
        self.pattern_regex = regex_pattern

    def resolve(self) -> List[FileInfo]:
        """Resolve files from S3 matching the given pattern."""

        objects = self.s3_client.list_objects_v2(Bucket=self.bucket)
        return [
            FileInfo(uri=f"s3://{self.bucket}/{obj['Key']}", last_modified=obj['LastModified'].timestamp())
            for obj in objects.get("Contents", [])
            if self.pattern_regex.fullmatch(obj['Key'])
        ]


class LocalFileResolver:
    def __init__(self, base_path: str, regex_pattern: re.Pattern):
        self.base_path = Path(base_path)
        self.pattern_regex = regex_pattern

    def resolve(self) -> List[FileInfo]:
        """Resolve local files based on regex pattern."""

        return [
            FileInfo(uri=f'file://{str(path)}', last_modified=path.stat().st_mtime)
            for path in self.base_path.rglob("*")
            if self.pattern_regex.fullmatch(str(path.relative_to(self.base_path))) and not path.is_dir()
        ]


def create_resolver(settings: Settings, base_path: str, regex_pattern: re.Pattern) -> FileResolver:
    """Factory function to create the appropriate resolver based on the URI template."""

    if base_path.startswith("s3://"):
        s3_client = get_s3_client(settings)
        parse_result = urlparse(base_path, allow_fragments=False)
        regex_pattern = re.compile(parse_result.path.lstrip('/') + regex_pattern.pattern)
        return S3FileResolver(s3_client=s3_client, bucket=parse_result.netloc, regex_pattern=regex_pattern)
    else:
        return LocalFileResolver(base_path=base_path, regex_pattern=regex_pattern)



import gzip
import os
import struct
import zipfile
from abc import ABC, abstractmethod
from io import BytesIO
from urllib.parse import unquote, urlparse

from scrapy.exceptions import NotConfigured
from scrapy.settings import Settings
from scrapy.utils.boto import is_botocore_available

from scrapy_webarchive.exceptions import UnsupportedURIException
from scrapy_webarchive.wacz import zip_utils


class ZipStorageHandler(ABC):
    """Abstract Base Class for ZIP storage handlers."""

    @abstractmethod
    def fetch_file(self, file_name: str) -> bytes:
        """Fetch the entire content of a file."""

        pass

    @abstractmethod
    def fetch_file_part(self, file_name: str, offset: int, size: int) -> bytes:
        """Fetch a specific part of a file."""

        pass

    @property
    @abstractmethod
    def zip_exists(self) -> bool:
        """Check if the ZIP file exists."""

        pass

    @property
    @abstractmethod
    def uri(self) -> str:
        """The URI of the ZIP file."""

        pass


class LocalZipStorageHandler(ZipStorageHandler):
    """Handles ZIP files stored locally."""

    def __init__(self, uri: str):
        """Initialize the handler with a local ZIP file."""

        self._uri = uri
        parsed_uri = urlparse(uri)
        self.file_path = unquote(parsed_uri.path)

        if not self.zip_exists:
            raise FileNotFoundError(f"ZIP file not found at {uri}")

        self.zip_file = zipfile.ZipFile(self.file_path)

    def fetch_file_part(self, file_name: str, offset: int, size: int) -> bytes:
        if file_name not in self.zip_file.namelist():
            raise FileNotFoundError(f"File {file_name} not found in ZIP archive")

        with self.zip_file.open(file_name) as file_obj:
            file_obj.seek(offset)

            if file_name.endswith(".gz"):
                with gzip.open(file_obj) as gz_file:
                    return gz_file.read()

            return file_obj.read()

    def fetch_file(self, file_name: str) -> bytes:
        if file_name not in self.zip_file.namelist():
            raise FileNotFoundError(f"File {file_name} not found in ZIP archive")

        with self.zip_file.open(file_name) as file_obj:
            if file_name.endswith(".gz"):
                with gzip.open(file_obj) as gz_file:
                    return gz_file.read()
            return file_obj.read()

    @property
    def zip_exists(self) -> bool:
        return os.path.exists(self.file_path)
    
    @property
    def uri(self) -> str:
        return self._uri


class RemoteZipStorageHandler(ZipStorageHandler):
    """Abstract base class for remote ZIP file handlers."""

    @abstractmethod
    def get_object(self, range_bytes: str) -> bytes:
        """Fetch an object (or part of it) from remote storage."""

        pass

    @abstractmethod
    def get_file_info(self) -> dict:
        """Get metadata about the remote file."""

        pass

    def _find_eocd(self, file_size: int) -> int:
        """Find the End of Central Directory (EOCD) in the ZIP file."""

        search_offset = max(0, file_size - 65536)
        range_bytes = f"bytes={search_offset}-{file_size - 1}"
        search_range = self.get_object(range_bytes)
        return zip_utils.find_eocd(search_range) + search_offset

    def _find_zip64_eocd(self, eocd_offset: int) -> int:
        """Find the ZIP64 End of Central Directory (EOCD)."""

        locator_offset = eocd_offset - 20
        range_bytes = f"bytes={locator_offset}-{eocd_offset - 1}"
        locator = self.get_object(range_bytes)
        return zip_utils.find_zip64_eocd(locator) + locator_offset

    def _get_file_header_length(self, offset: int) -> int:
        """Retrieve the length of a file header from the ZIP file."""

        range_bytes = f"bytes={offset}-{offset + 29}"
        local_header = self.get_object(range_bytes)

        if local_header[:4] != b"\x50\x4b\x03\x04":
            raise ValueError("Invalid Local File Header signature")

        file_name_length, extra_field_length = struct.unpack("<HH", local_header[26:30])
        return 30 + file_name_length + extra_field_length

    def _get_zip_metadata(self) -> dict:
        """Retrieve metadata from the ZIP file, including central directory information."""

        file_size = self.get_file_info()["ContentLength"]
        eocd_offset = self._find_eocd(file_size)
        range_bytes = f"bytes={eocd_offset}-{eocd_offset + zip_utils.EOCD_RECORD_SIZE - 1}"
        eocd = self.get_object(range_bytes)

        if zip_utils.is_zip64(eocd):
            zip64_eocd_offset = self._find_zip64_eocd(eocd_offset)
            range_bytes = f"bytes={zip64_eocd_offset}-{zip64_eocd_offset + zip_utils.ZIP64_EOCD_RECORD_SIZE - 1}"
            zip64_eocd = self.get_object(range_bytes)
            cd_start, cd_size = zip_utils.parse_zip64_eocd(zip64_eocd)
        else:
            cd_start, cd_size = zip_utils.parse_eocd(eocd)

        range_bytes = f"bytes={cd_start}-{cd_start + cd_size - 1}"
        central_directory = self.get_object(range_bytes)

        return zip_utils.parse_central_directory(central_directory, self._get_file_header_length)


class S3ZipStorageHandler(RemoteZipStorageHandler):
    """Handles ZIP files stored in Amazon S3."""

    def __init__(self, uri: str, s3_client):
        if not is_botocore_available():
            raise NotConfigured("missing botocore library")

        parse_result = urlparse(uri, allow_fragments=False)
        if parse_result.scheme != "s3":
            raise ValueError(f"Incorrect URI scheme in {uri}, expected 's3'")

        self._uri = uri
        self.bucket = parse_result.netloc
        self.path = parse_result.path.lstrip("/")
        self.s3_client = s3_client
        self.zip_metadata = self._get_zip_metadata()

    def get_object(self, range_bytes: str) -> bytes:
        response = self.s3_client.get_object(Bucket=self.bucket, Key=self.path, Range=range_bytes)
        return response["Body"].read()

    def get_file_info(self) -> dict:
        return self.s3_client.head_object(Bucket=self.bucket, Key=self.path)

    def fetch_file_part(self, file_name: str, offset: int, size: int) -> bytes:
        metadata = self.zip_metadata[file_name]
        range_bytes = self._calculate_range(metadata, offset, size)
        data = self.get_object(range_bytes)

        return gzip.open(BytesIO(data)).read() if file_name.endswith(".gz") else data

    def fetch_file(self, file_name: str) -> bytes:
        metadata = self.zip_metadata[file_name]
        range_bytes = self._calculate_range(metadata, 0, metadata["compressed_size"])
        return self.get_object(range_bytes)

    def _calculate_range(self, metadata: dict, offset: int, size: int) -> str:
        start = metadata["header_offset"] + metadata["file_header_length"] + offset
        end = start + size - 1
        return f"bytes={start}-{end}"

    @property
    def zip_exists(self) -> bool:
        return bool(self.get_file_info())

    @property
    def uri(self) -> str:
        return self._uri


class ZipStorageHandlerFactory:
    @staticmethod
    def get_handler(uri: str, settings: Settings) -> ZipStorageHandler:
        if uri.startswith('s3://'):
            s3_client = get_s3_client(settings)
            return S3ZipStorageHandler(uri, s3_client)
        elif uri.startswith('file://'):
            return LocalZipStorageHandler(uri)
        else:
            raise UnsupportedURIException(f"URI scheme not supported: {uri}")


def get_s3_client(settings: Settings):
    import botocore.session
    session = botocore.session.get_session()
    return session.create_client(
        "s3",
        aws_access_key_id=settings["AWS_ACCESS_KEY_ID"],
        aws_secret_access_key=settings["AWS_SECRET_ACCESS_KEY"],
        aws_session_token=settings["AWS_SESSION_TOKEN"],
        endpoint_url=settings["AWS_ENDPOINT_URL"],
        region_name=settings["AWS_REGION_NAME"],
        use_ssl=settings["AWS_USE_SSL"],
        verify=settings["AWS_VERIFY"],
    )



from __future__ import annotations

import os
from io import BytesIO

from scrapy import Spider, signals
from scrapy.crawler import Crawler
from scrapy.exceptions import NotConfigured
from scrapy.http.request import Request
from scrapy.http.response import Response
from scrapy.pipelines.files import FSFilesStore, FTPFilesStore, GCSFilesStore, S3FilesStore
from scrapy.pipelines.media import MediaPipeline
from scrapy.settings import Settings
from twisted.internet.defer import Deferred
from typing_extensions import Any, Dict, Protocol, Self, Tuple, Type, Union, cast

from scrapy_webarchive.constants import WARC_DT_FORMAT, WEBARCHIVE_META_KEY
from scrapy_webarchive.models import WarcMetadata
from scrapy_webarchive.utils import (
    get_archive_uri_template_dt_variables,
    get_formatted_dt_string,
    get_scheme_from_uri,
    is_uri_directory,
)
from scrapy_webarchive.wacz.creator import WaczFileCreator
from scrapy_webarchive.warc.writers import WarcFileWriter


class FilesStoreProtocol(Protocol):
    # This protocol will be introduced by Scrapy at a later stage. Once that happens we can drop this bit.
    # For now we implement it ourselves to improve type hints for the stores.
    def __init__(self, basedir: str): 
        ...

    def persist_file(
        self,
        path: str,
        buf: BytesIO,
        info: MediaPipeline.SpiderInfo,
        meta: Union[Dict[str, Any], None] = None,
        headers: Union[Dict[str, str], None] = None,
    ) -> Union[Deferred[Any], None]: 
        ...


class WaczExporter:
    """WACZ exporter extension that writes spider requests/responses as WARC and later compiles them to a WACZ."""

    wacz_fname = None
    STORE_SCHEMES: Dict[str, Type[FilesStoreProtocol]] = {
        "": FSFilesStore,
        "file": FSFilesStore,
        "s3": S3FilesStore,
        "gs": GCSFilesStore,
        "ftp": FTPFilesStore,
    }

    def __init__(self, settings: Settings, crawler: Crawler) -> None:
        self.settings = settings
        
        # Check configuration prerequisites
        self._check_configuration_prerequisites()

        self.stats = crawler.stats
        self.crawler = crawler
        self.spider_name = crawler.spidercls.name if hasattr(crawler.spidercls, "name") else crawler.spider.name

        # Get the store URI and configure the WACZ filename
        self.store_uri, self.wacz_fname  = self._retrieve_store_uri_and_wacz_fname()

        # Initialize store, writer and creator
        self.store: FilesStoreProtocol = self._get_store(self.store_uri)
        self.writer = WarcFileWriter(collection_name=self.spider_name)
        self.wacz_creator = WaczFileCreator(
            store=self.store,
            warc_fname=self.writer.warc_fname,
            wacz_fname=self.wacz_fname,
            collection_name=self.spider_name,
            title=self.settings["SW_WACZ_TITLE"],
            description=self.settings["SW_WACZ_DESCRIPTION"],
        )

    def _check_configuration_prerequisites(self) -> None:
        """raises NotConfigured if essential settings or middleware configurations are incorrect."""
        
        if not self.settings.get("SW_EXPORT_URI"):
            raise NotConfigured("Missing SW_EXPORT_URI setting.")
        
        if any(
            [
                self.settings.get("SW_WACZ_SOURCE_URI"),
                self.settings.get("SW_WACZ_LOOKUP_STRATEGY"),
                self.settings.get("SW_WACZ_LOOKUP_TARGET"),
            ]
        ):
            raise NotConfigured("WACZ exporter is disabled when scraping from a WACZ archive.")

    def _retrieve_store_uri_and_wacz_fname(self) -> Tuple[str, Union[str, None]]:
        """Sets up the export URI based on configuration and spider context."""

        export_uri = self.settings["SW_EXPORT_URI"].format(
            spider=self.spider_name,
            **get_archive_uri_template_dt_variables(),
        )

        if is_uri_directory(export_uri):
            return export_uri, None
        else:
            export_uri, wacz_fname = os.path.split(export_uri)
            return f"{export_uri}/", wacz_fname

    def _get_store(self, store_uri: str) -> FilesStoreProtocol:
        store_cls = self.STORE_SCHEMES[get_scheme_from_uri(store_uri)]
        return store_cls(store_uri)

    @classmethod
    def from_crawler(cls, crawler: Crawler) -> Self:
        assert crawler.stats

        try:
            exporter = cls.from_settings(crawler.settings, crawler)
        except AttributeError:
            exporter = cls(crawler.settings, crawler)

        crawler.signals.connect(exporter.response_downloaded, signal=signals.response_downloaded)
        crawler.signals.connect(exporter.spider_closed, signal=signals.spider_closed)
        crawler.signals.connect(exporter.spider_opened, signal=signals.spider_opened)
        return exporter
    
    @classmethod
    def from_settings(cls, settings: Settings, crawler: Crawler):
        """
        Store configuration is based on the images/files pipeline from Scrapy. See:
        https://github.com/scrapy/scrapy/blob/d4709e41047e794c9e39968f61c5abbcddf825c4/scrapy/pipelines/images.py#L94-L116
        """

        s3store: Type[S3FilesStore] = cast(Type[S3FilesStore], cls.STORE_SCHEMES["s3"])
        s3store.AWS_ACCESS_KEY_ID = settings["AWS_ACCESS_KEY_ID"]
        s3store.AWS_SECRET_ACCESS_KEY = settings["AWS_SECRET_ACCESS_KEY"]
        s3store.AWS_SESSION_TOKEN = settings["AWS_SESSION_TOKEN"]
        s3store.AWS_ENDPOINT_URL = settings["AWS_ENDPOINT_URL"]
        s3store.AWS_REGION_NAME = settings["AWS_REGION_NAME"]
        s3store.AWS_USE_SSL = settings["AWS_USE_SSL"]
        s3store.AWS_VERIFY = settings["AWS_VERIFY"]
        s3store.POLICY = settings["FILES_STORE_S3_ACL"]

        gcs_store: Type[GCSFilesStore] = cast(Type[GCSFilesStore], cls.STORE_SCHEMES["gs"])
        gcs_store.GCS_PROJECT_ID = settings["GCS_PROJECT_ID"]
        gcs_store.POLICY = settings["FILES_STORE_GCS_ACL"] or None

        ftp_store: Type[FTPFilesStore] = cast(Type[FTPFilesStore], cls.STORE_SCHEMES["ftp"])
        ftp_store.FTP_USERNAME = settings["FTP_USER"]
        ftp_store.FTP_PASSWORD = settings["FTP_PASSWORD"]
        ftp_store.USE_ACTIVE_MODE = settings.getbool("FEED_STORAGE_FTP_ACTIVE")

        return cls(settings=settings, crawler=crawler)

    def spider_opened(self) -> None:
        self.writer.write_warcinfo(robotstxt_obey=self.settings["ROBOTSTXT_OBEY"])

    def response_downloaded(self, response: Response, request: Request, spider: Spider) -> None:
        request.meta["WARC-Date"] = get_formatted_dt_string(format=WARC_DT_FORMAT)

        # Write response WARC record
        response_record = self.writer.write_response(response, request)
        self.stats.inc_value("webarchive/exporter/response_written", spider=spider)
        self.stats.inc_value(
            f"webarchive/exporter/writer_status_count/{response_record.http_headers.get_statuscode()}", 
            spider=spider,
        )

        # Write request WARC record
        self.writer.write_request(request, concurrent_to=response_record)
        self.stats.inc_value("webarchive/exporter/request_written", spider=spider)

        warc_metadata = WarcMetadata(
            action="write",
            record_id=response_record.rec_headers.get_header('WARC-Record-ID'),
            wacz_uri=self.export_uri,
        )
        request.meta[WEBARCHIVE_META_KEY] = warc_metadata.to_dict()

    def spider_closed(self, spider: Spider) -> None:
        self.wacz_creator.create()

    @property
    def export_uri(self) -> str:
        return os.path.join(self.store_uri, self.wacz_creator.wacz_fname)



from dataclasses import dataclass

from scrapy.http.response import Response
from typing_extensions import Union

from scrapy_webarchive.constants import WEBARCHIVE_META_KEY


@dataclass
class WarcMetadata:
    """
    Encapsulates metadata about the WARC record.

    Attributes:
        action (str): The action performed ("read" or "write").
        record_id (str): The unique ID of the WARC record.
        wacz_uri (str): The URI of the WACZ file.
    """
    action: str
    record_id: str
    wacz_uri: str

    def to_dict(self) -> dict:
        """Convert the object to a dictionary for compatibility with Scrapy's meta."""

        return {
            "action": self.action,
            "record_id": self.record_id,
            "wacz_uri": self.wacz_uri,
        }

    @classmethod
    def from_response(cls, response: Response) -> Union["WarcMetadata", None]:
        """Create a WarcMetadata instance from a Scrapy response object."""

        if not hasattr(response, "meta"):
            return None

        warc_meta = response.meta.get(WEBARCHIVE_META_KEY)

        if not warc_meta:
            return None

        return cls(**warc_meta)


@dataclass
class FileInfo:
    uri: str
    last_modified: float

    def __repr__(self):
        return f"FileInfo(uri={self.uri}, last_modified={self.last_modified})"

    def __lt__(self, other):
        if isinstance(other, FileInfo):
            return self.last_modified < other.last_modified
        return NotImplemented



from unittest import mock

import pytest
from scrapy.exceptions import NotConfigured
from scrapy.http import Request, Response
from scrapy.pipelines.files import FSFilesStore, FTPFilesStore, GCSFilesStore, S3FilesStore
from scrapy.utils.test import get_crawler

from scrapy_webarchive.extensions import WaczExporter


class TestWaczExporterExtension:
    def test_archive_export_uri_invalid_raises_not_configured(self):
        crawler = get_crawler(settings_dict={})
        with pytest.raises(NotConfigured):
            WaczExporter.from_crawler(crawler)

    @mock.patch('scrapy_webarchive.extensions.S3FilesStore.__init__', return_value=None)
    @mock.patch('scrapy_webarchive.extensions.GCSFilesStore.__init__', return_value=None)
    @mock.patch('scrapy_webarchive.extensions.FTPFilesStore.__init__', return_value=None)
    @mock.patch('scrapy_webarchive.extensions.FSFilesStore.__init__', return_value=None)
    def test_get_store(self, *args):
        crawler = get_crawler(settings_dict={"SW_EXPORT_URI": "/tmp/scrapy-webarchive/wacz/"})
        crawler.spider = crawler._create_spider("quotes")
        extension = WaczExporter.from_crawler(crawler)
        assert isinstance(extension.store, FSFilesStore)
    
        crawler = get_crawler(settings_dict={"SW_EXPORT_URI": "s3://scrapy-webarchive/wacz/"})
        crawler.spider = crawler._create_spider("quotes")
        extension = WaczExporter.from_crawler(crawler)
        assert isinstance(extension.store, S3FilesStore)

        crawler = get_crawler(settings_dict={"SW_EXPORT_URI": "gs://scrapy-webarchive/wacz/"})
        crawler.spider = crawler._create_spider("quotes")
        extension = WaczExporter.from_crawler(crawler)
        assert isinstance(extension.store, GCSFilesStore)

        crawler = get_crawler(settings_dict={"SW_EXPORT_URI": "ftp://scrapy-webarchive/wacz/"})
        crawler.spider = crawler._create_spider("quotes")
        extension = WaczExporter.from_crawler(crawler)
        assert isinstance(extension.store, FTPFilesStore)

    def test_response_downloaded(self):
        crawler = get_crawler(settings_dict={"SW_EXPORT_URI": "/tmp/scrapy-webarchive/wacz/"})
        crawler.spider = crawler._create_spider("quotes")
        extension = WaczExporter.from_crawler(crawler)
        extension.writer = mock.Mock()

        # Call the method under test
        request = Request("http://example.com")
        response = Response(request.url)
        extension.response_downloaded(response, request, crawler.spider)

        # Verify that the WARC date was set in request meta
        assert "WARC-Date" in request.meta

        # Verify that the response and request were written to the WARC file
        extension.writer.write_response.assert_called_once()
        extension.writer.write_request.assert_called_once()

        # Verify that the stats were incremented correctly
        assert extension.stats._stats['webarchive/exporter/response_written'] == 1
        assert extension.stats._stats['webarchive/exporter/request_written'] == 1



from datetime import datetime

import pytest
from typing_extensions import List

from scrapy_webarchive.models import FileInfo
from scrapy_webarchive.strategies import AfterStrategy, BeforeStrategy, StrategyRegistry


@pytest.fixture
def sample_files() -> List[FileInfo]:
    return [
        FileInfo("archive_1.wacz", datetime(2025, 1, 1).timestamp()),
        FileInfo("archive_2.wacz", datetime(2025, 6, 1).timestamp()),
        FileInfo("archive_3.wacz", datetime(2025, 12, 1).timestamp()),
    ]


@pytest.mark.parametrize("strategy_name", ["before", "after"])
def test_strategy_registry_register_defaults(strategy_name):
    assert strategy_name in StrategyRegistry._strategies


@pytest.mark.parametrize("strategy_name, strategy_cls", [
    ("before", BeforeStrategy),
    ("after", AfterStrategy),
])
def test_strategy_registry_get_valid(strategy_name, strategy_cls):
    strategy = StrategyRegistry.get(strategy_name)
    assert isinstance(strategy, strategy_cls)


def test_strategy_registry_get_invalid():
    with pytest.raises(ValueError, match="Unknown strategy: invalid"):
        StrategyRegistry.get("invalid")


def test_strategy_registry_new_registration():
    StrategyRegistry._strategies.clear()
    
    class NewStrategy():
        pass
    
    StrategyRegistry.register("new")(NewStrategy)
    assert StrategyRegistry._strategies["new"] == NewStrategy


@pytest.mark.parametrize(
    "strategy, target_date, files_input, expected_result",
    [
        (BeforeStrategy(), datetime(2025, 7, 1), "sample_files", "archive_2.wacz"),
        (BeforeStrategy(), datetime(2024, 12, 1), "sample_files", None),
        (BeforeStrategy(), datetime(2025, 7, 1), [], None),
        (BeforeStrategy(), datetime(2025, 6, 1), "sample_files", "archive_2.wacz"),
        (AfterStrategy(), datetime(2025, 3, 1), "sample_files", "archive_2.wacz"),
        (AfterStrategy(), datetime(2026, 1, 1), "sample_files", None),
        (AfterStrategy(), datetime(2025, 7, 1), [], None),
        (AfterStrategy(), datetime(2025, 6, 1), "sample_files", "archive_2.wacz"),
    ],
    ids=[
        "before_finds_file_before_target",
        "before_no_file_before_target",
        "before_empty_list",
        "before_exact_match",
        "after_finds_file_after_target",
        "after_no_file_after_target",
        "after_empty_list",
        "after_exact_match",
    ]
)
def test_strategy_find(strategy, target_date, files_input, expected_result, sample_files):
    files = sample_files if files_input == "sample_files" else files_input
    result = strategy.find(files, target_date)

    if isinstance(result, dict):
        result = result.get('name')

    assert result == expected_result



from __future__ import annotations

import importlib
import os
from datetime import datetime
from importlib.util import find_spec

from scrapy.utils.project import ENVVAR
from typing_extensions import Dict, List, Optional, Protocol, Type

from scrapy_webarchive.models import FileInfo


class FileLookupStrategy(Protocol):
    def find(self, files: List[FileInfo], target: datetime) -> Optional[str]:
        """Algorithm to find a file based on a target."""
        ...


class StrategyRegistry:
    _strategies: Dict[str, Type[FileLookupStrategy]] = {}

    @classmethod
    def register(cls, name: str):
        """Decorator to register a new strategy."""

        def decorator(strategy_cls: Type[FileLookupStrategy]):
            cls._strategies[name] = strategy_cls
            return strategy_cls  # Ensure the class is returned for normal usage
        return decorator

    @classmethod
    def get(cls, name: str) -> FileLookupStrategy:
        """Retrieves a strategy instance by name."""

        if name not in cls._strategies:
            raise ValueError(f"Unknown strategy: {name}")
        return cls._strategies[name]()
    
    @classmethod
    def auto_discover(cls):
        """Auto-discovers strategies from the current Scrapy project."""

        scrapy_module = os.environ.get(ENVVAR)

        if not scrapy_module:
            return

        project_root = scrapy_module.rsplit(".", 1)[0]
        strategies_module = f"{project_root}.strategies"

        if find_spec(strategies_module):
            importlib.import_module(strategies_module)


@StrategyRegistry.register("before")
class BeforeStrategy:
    """Strategy to find the file that was last modified before the target time. Should match closest to the target."""

    def find(self, files: List[FileInfo], target: datetime) -> Optional[str]:
        sorted_files = sorted(files, reverse=True)
        target_timestamp = target.timestamp()

        for file in sorted_files:
            if target_timestamp >= file.last_modified:
                return file.uri

        return None


@StrategyRegistry.register("after")
class AfterStrategy:
    """Strategy to find the file that was last modified after the target time. Should match closest to the target."""

    def find(self, files: List[FileInfo], target: datetime) -> Optional[str]:
        sorted_files = sorted(files)
        target_timestamp = target.timestamp()

        for file in sorted_files:
            if file.last_modified >= target_timestamp:
                return file.uri

        return None


StrategyRegistry.auto_discover()



from scrapy.http.request import Request
from scrapy.responsetypes import ResponseTypes
from warc.warc import WARCRecord

from scrapy_webarchive.cdxj.models import CdxjRecord
from scrapy_webarchive.exceptions import WaczMiddlewareException
from scrapy_webarchive.utils import header_lines_to_dict


class WarcRecordTransformer:
    """A helper class to convert WARC records into Scrapy requests and responses."""
    
    response_types = ResponseTypes()

    def request_for_record(self, cdxj_record: CdxjRecord, **kwargs):
        """Create a Scrapy request instance from a WARCRecord."""

        # TODO: locate request in WACZ and include all relevant things (like headers)
        return Request(url=cdxj_record.data["url"], method=cdxj_record.data.get("method", "GET"), **kwargs)

    def response_for_record(self, warc_record: WARCRecord, request: Request, **kwargs):
        """Create a Scrapy response instance from a WARCRecord."""

        # We expect a response.
        # https://iipc.github.io/warc-specifications/specifications/warc-format/warc-1.1/#warc-type-mandatory
        if warc_record.type != "response":
            raise WaczMiddlewareException(f"Unexpected record type: {warc_record.type}")

        # We only know how to handle application/http.
        # https://iipc.github.io/warc-specifications/specifications/warc-format/warc-1.1/#content-type
        record_content_type = (warc_record["Content-Type"] or "").split(";", 1)[0]
        if record_content_type != "application/http":
            raise WaczMiddlewareException(f"Unexpected record content-type: {record_content_type}")

        # There is a date field in record['WARC-Date'], but don't have a use for it now.
        # https://iipc.github.io/warc-specifications/specifications/warc-format/warc-1.1/#warc-date-mandatory

        payload = warc_record.payload.read()
        payload_parts = payload.split(b"\r\n\r\n", 1)
        header_lines = payload_parts[0] if len(payload_parts) > 0 else ""
        body = payload_parts[1] if len(payload_parts) > 1 else None

        header_lines = header_lines.split(b"\r\n")
        header_parts = header_lines[0].split(None, 2)
        protocol = header_parts[0] if len(header_parts) > 0 else None
        status = header_parts[1] if len(header_parts) > 1 else None
        headers = header_lines_to_dict(header_lines[1:])

        if not status or not protocol:
            return None

        response_cls = self.response_types.from_headers(headers)

        return response_cls(
            url=warc_record.url,
            request=request,
            status=int(status.decode()),
            protocol=protocol.decode(),
            headers=headers,
            body=body,
            **kwargs,
        )


record_transformer = WarcRecordTransformer()



from __future__ import annotations

from cdxj_indexer.main import CDXJIndexer
from typing_extensions import List


def write_cdxj_index(output: str, inputs: List[str]) -> str:
    """Generates a CDXJ index from a list of input WARC files and writes the index to an output file."""

    wacz_indexer = CDXJIndexer(output=output, inputs=inputs)
    wacz_indexer.process_all()
    return output



from dataclasses import asdict

import pytest
from scrapy.http import Request, Response

from scrapy_webarchive.constants import WEBARCHIVE_META_KEY
from scrapy_webarchive.models import FileInfo, WarcMetadata


class TestWarcMetadata:
    def setup_method(self):
        self.meta = {
            "action": "read",
            "record_id": "12345",
            "wacz_uri": "s3://example/archive.wacz",
        }

    def test_warc_metadata_to_dict(self):
        result = WarcMetadata(**self.meta).to_dict()
        assert result == self.meta

    def test_warc_metadata_from_response_valid(self):
        warc_metadata = WarcMetadata.from_response(self._get_response())
        assert isinstance(warc_metadata, WarcMetadata)
        assert warc_metadata.action == "read"
        assert warc_metadata.record_id == "12345"
        assert warc_metadata.wacz_uri == "s3://example/archive.wacz"

    def test_warc_metadata_from_response_no_meta(self):
        response = self._get_response(meta={})
        warc_meta = WarcMetadata.from_response(response)
        assert warc_meta is None

    def test_warc_metadata_from_response_no_warc_meta(self):
        response = self._get_response(meta={"key": "value"})
        warc_meta = WarcMetadata.from_response(response)
        assert warc_meta is None

    def test_warc_metadata_from_response_invalid_meta(self):
        response = self._get_response(meta={WEBARCHIVE_META_KEY: {"invalid_key": "value"}})
        with pytest.raises(TypeError):
            WarcMetadata.from_response(response)
    
    def _get_response(self, meta: dict = None, url: str = "https://example.com"):
        if meta is None:
            meta = {WEBARCHIVE_META_KEY: self.meta}

        request = Request(url=url, meta=meta)
        return Response(url=url, request=request)


def test_fileinfo_creation():
    file_info = FileInfo(uri="archive.wacz", last_modified=123.456)
    assert file_info.uri == "archive.wacz"
    assert file_info.last_modified == 123.456

    dict_repr = asdict(file_info)
    assert dict_repr == {"uri": "archive.wacz", "last_modified": 123.456}


def test_fileinfo_repr():
    file_info = FileInfo(uri="archive.wacz", last_modified=123.456)
    expected_repr = "FileInfo(uri=archive.wacz, last_modified=123.456)"
    assert repr(file_info) == expected_repr


def test_fileinfo_comparison_lt():
    file1 = FileInfo(uri="file1.txt", last_modified=123.456)
    file2 = FileInfo(uri="file2.txt", last_modified=789.123)
    
    assert file1 < file2
    assert not file2 < file1


def test_fileinfo_comparison_same_timestamp():
    file1 = FileInfo(uri="file1.txt", last_modified=123.456)
    file2 = FileInfo(uri="file2.txt", last_modified=123.456)
    
    assert not file1 < file2
    assert not file2 < file1


def test_fileinfo_comparison_invalid_type():
    file_info = FileInfo(uri="archive.wacz", last_modified=123.456)
    result = file_info.__lt__("invalid_type")
    assert result == NotImplemented
